// scalar product: 4: use of cublasDdot
#include <cassert>
#include <iostream>
#include <ctime>
using namespace std;
#include <cublas_v2.h>   //  new API, see https://docs.nvidia.com/cuda/cublas/index.html#new-and-legacy-cublas-api

int main(void)         //  main function on host
{
    const int N = 100000000;
    //const int N = 14000000;
    const int LOOP = 100;

    float *x_h = new float [N]; // host data
    float *y_h = new float [N]; // host data

    float ss = 0.0;
    for (int i = 0; i < N; i++) {
        x_h[i] = (i % 137) + 1;
        y_h[i] = 1.0 / x_h[i];
        ss += x_h[i] * y_h[i];
    }
//  -- Initialize cuBLAS, check GPU
//  https://developer.nvidia.com/sites/default/files/akamai/cuda/files/Misc/mygpu.pdf
    cudaError_t  cudaStat;               //  cudaMalloc  status
    cublasHandle_t  handle;

    cublasStatus_t stat = cublasCreate (& handle );   //  initialize  CUBLAS  context
    if (stat != CUBLAS_STATUS_SUCCESS) {
        fprintf(stderr, "ERROR: cublasInit() failed!\n");
        exit(1);
    }

    cudaDeviceProp prop;
    cudaGetDeviceProperties (&prop, 0);
    cout << endl << "We work on a " << prop.name << " GPU  with " <<
         prop.multiProcessorCount << " Multiprocessors" << endl;

//  -- Move data from host to device
// allocate memory on device
    float *x_d, *y_d; // device data
    cudaMalloc((void **) &x_d, N * sizeof(float));
    cudaMalloc((void **) &y_d, N * sizeof(float));

// copy data:  host --> device
    cublasSetVector(N, sizeof(float), x_h, 1, x_d, 1);
    cublasSetVector(N, sizeof(float), y_h, 1, y_d, 1);

//  -- Benchmark
    float sum = 0.0f;
    float res;
    cudaDeviceSynchronize();
    double tstart = clock();
    for (int k = 0; k < LOOP; ++k) {
        stat = cublasSdot(handle, N, x_d, 1, y_d, 1, &res);
        sum += res;
    }
    cudaDeviceSynchronize();
    double tgpu = (clock() - tstart) / CLOCKS_PER_SEC / LOOP;

//  -- Statistics, check result
    cout << "GPU: " << tgpu << " sec.; " <<
         2e-9 * N / tgpu << " GFLOPS;  " <<
         2.0 * N *sizeof(float) / tgpu / 1024 / 1024 / 1024 << " GiB/sec.\n ";

    //if (static_cast<int>(sum / LOOP) != N) {      // round-off errors with 'float'
    if (static_cast<int>(res) != N) {
        cout << " !!  W R O N G  result  " << endl;
        cout << res << "  vs.  " << N << "(" << ss << ")" << endl;
    }
    else
        cout << " correct  result  " << endl;

//  -- Free all cuBlas ressources
    cudaFree(x_d);
    cudaFree(y_d);
    cublasDestroy(handle );

//  -- free host ressources
    delete [] x_h;
    delete [] y_h;

    return 0;
}
